/**
 *
 * Copyright (C) 2014 Olivier Berthier
 *                    http://www.xunil.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include "Bounding.h"

BBox::BBox(const Coord& p1, const Coord& p2){
  empty=0;
  min = Coord( 
	      std::min( p1.getX(), p2.getX() ),
	      std::min( p1.getY(), p2.getY() ),
	      std::min( p1.getZ(), p2.getZ() ) );
        
  max = Coord( 
	      std::max( p1.getX(), p2.getX() ),
	      std::max( p1.getY(), p2.getY() ),
	      std::max( p1.getZ(), p2.getZ() ) );

}

void BBox::add(const Coord& p){
  if(empty){
    min = p;
    max = p;
    empty=0;
  }else{
    if(p.getX() < min.getX())
      min.setX( p.getX());
    if(p.getY() < min.getY())
      min.setY( p.getY());
    if(p.getZ() < min.getZ())
      min.setZ( p.getZ());
        
    if(p.getX() > max.getX())
      max.setX( p.getX());
    if(p.getY() > max.getY())
      max.setY( p.getY());
    if(p.getZ() > max.getZ())
      max.setZ( p.getZ());
  }
  // min.Print();
  // max.Print();
}

void BBox::draw() const{
  glColor3f( 1.0, 0.0, 0.0 ); 
  glBegin(GL_POLYGON);  
  glVertex3f( max.getX(), min.getY(), min.getZ() ); 
  glVertex3f( max.getX(),  max.getY(), min.getZ() );
  glVertex3f( min.getX(),  max.getY(), min.getZ() );
  glVertex3f( min.getX(), min.getY(), min.getZ() ); 
  glEnd();
 
  // White side - BACK
  glBegin(GL_POLYGON);
  glVertex3f( max.getX(), min.getY(), max.getZ() );
  glVertex3f( max.getX(),  max.getY(), max.getZ() );
  glVertex3f( min.getX(),  max.getY(), max.getZ() );
  glVertex3f( min.getX(), min.getY(), max.getZ() );
  glEnd();
 
  // Purple side - RIGHT
  glBegin(GL_POLYGON);
  glVertex3f( max.getX(), min.getY(), min.getZ() );
  glVertex3f( max.getX(),  max.getY(), min.getZ() );
  glVertex3f( max.getX(),  max.getY(),  max.getZ() );
  glVertex3f( max.getX(), min.getY(),  max.getZ() );
  glEnd();
 
  // Green side - LEFT
  glBegin(GL_POLYGON);
  glVertex3f( min.getX(), min.getY(),  max.getZ() );
  glVertex3f( min.getX(),  max.getY(),  max.getZ() );
  glVertex3f( min.getX(),  max.getY(), min.getZ() );
  glVertex3f( min.getX(), min.getY(), min.getZ() );
  glEnd();
 
  // Blue side - TOP
  glBegin(GL_POLYGON);
  glVertex3f(  max.getX(),  max.getY(),  max.getZ() );
  glVertex3f(  max.getX(),  max.getY(), min.getZ() );
  glVertex3f( min.getX(),  max.getY(), min.getZ() );
  glVertex3f( min.getX(),  max.getY(),  max.getZ() );
  glEnd();
 
  // Red side - BOTTOM
  glBegin(GL_POLYGON);
  glVertex3f(  max.getX(), min.getY(), min.getZ() );
  glVertex3f(  max.getX(), min.getY(),  max.getZ() );
  glVertex3f( min.getX(), min.getY(),  max.getZ() );
  glVertex3f( min.getX(), min.getY(), min.getZ() );
  glEnd();
 

}

int BBox::inside(const Coord& p){
  // if( p.getX() >= min.getX() && p.getX() <= max.getX() 
  //     && p.getY() >= min.getY() && p.getY() <= max.getY() 
  //     && p.getZ() >= min.getZ() && p.getZ() <= max.getZ() ){
  //   std::cout<<"====="<<std::endl;
  //   Coord a(p);
  //   a.Print();
  //   std::cout<<"in"<<std::endl;
  //   min.Print();
  //   max.Print();
  //  }
  return ( p.getX() >= min.getX() && p.getX() <= max.getX() 
	   && p.getY() >= min.getY() && p.getY() <= max.getY() 
	   && p.getZ() >= min.getZ() && p.getZ() <= max.getZ() );
}
int BBox::intersect(const BBox& box){
  int x = ( max.getX() >= box.min.getX() ) && ( min.getX() <= box.max.getX() );
  int y = ( max.getY() >= box.min.getY() ) && ( min.getY() <= box.max.getY() );
  int z = ( max.getZ() >= box.min.getZ() ) && ( min.getZ() <= box.max.getZ() );
  return ( x && y && z );
}

 


void tree::build() {
  
  for (std::vector<Coord>::iterator itr = objects.begin(); itr < objects.end(); itr++) {
    Coord curr = (*itr);
    bounds.add(curr);
  }

  if (objects.size() <= 20) 
    return;
  
  double splitValue=0;
  switch(axis) {
  case 0: 
    splitValue = (bounds.min.getX() + bounds.max.getX()) / 2.0;
  case 1: 
    splitValue = (bounds.min.getY() + bounds.max.getY()) / 2.0;
  case 2: 
    splitValue = (bounds.min.getZ() + bounds.max.getZ()) / 2.0;
  }  

  //remplacer par patition...
  std::vector<Coord> leftObjects;
  std::vector<Coord> rightObjects;

  for (std::vector<Coord>::iterator itr = objects.begin(); itr < objects.end(); itr++) {
    Coord curr = *itr;


    double pos;
    switch(axis) {
    case 0:
      pos = curr.getX();
      break;
    case 1:
      pos = curr.getY();
      break;
    case 2:
      pos = curr.getZ();
      break;
    }

    if (pos < splitValue) {
      leftObjects.push_back(curr);
    }else {
      rightObjects.push_back(curr);
    }
  }

  //  std::cout<<objects.size()<<" => "<<leftObjects.size()<<" + "<<rightObjects.size()<<std::endl;
  int newAxis = toggleAxis();

  if (leftObjects.size() != objects.size() && rightObjects.size() != objects.size()) {
    left = new tree(depth + 1, newAxis, leftObjects);
    right = new tree(depth + 1, newAxis, rightObjects);
  } else if (axisRetries < 2) {
    axis = toggleAxis();
    axisRetries++;
    build();
  } 
}

int tree::toggleAxis() {
  if(axis==1) return 2;
  else if(axis == 2) return 0;
  else return 1;
}




// int BBox::intersect(const BSphere& bs){
//   float dist_squared = bs.getRayon() * bs.getRayon();

//   if (bs.getCenter().getX() < min.getX()) 
//     dist_squared -= (bs.getCenter().getX() - min.getX())*(bs.getCenter().getX() - min.getX());
//   else if (bs.getCenter().getX() > max.getX())
//     dist_squared -= (bs.getCenter().getX() - max.getX())*(bs.getCenter().getX() - max.getX());
//   if (bs.getCenter().getY() < min.getY())
//     dist_squared -= (bs.getCenter().getY() - min.getY())*(bs.getCenter().getY() - min.getY());
//   else if (bs.getCenter().getY() > max.getY()) 
//     dist_squared -= (bs.getCenter().getY() - max.getY())*(bs.getCenter().getY() - max.getY());
//   if (bs.getCenter().getZ() < min.getZ())
//     dist_squared -= (bs.getCenter().getZ() - min.getZ())*(bs.getCenter().getZ() - min.getZ());
//   else if (bs.getCenter().getZ() > max.getZ())
//     dist_squared -= (bs.getCenter().getZ() - max.getZ())*(bs.getCenter().getZ() - max.getZ());
//   return dist_squared > 0;
// }

// BSphere::BSphere(const Coord& _position, float _rayon): position(_position),rayon(_rayon){}

// int BSphere::inside(const Coord& p) const{
//   Coord a(p.getX() - position.getX(),p.getY() - position.getY(),p.getZ() - position.getZ()) ;
//   float distance = Norm(a);
//   if(distance <= rayon){
//     if(left != NULL || right != NULL){
//       return left->inside(p) || right->inside(p);	
//     }
//   }
//   return 0;
// }

// int BSphere::intersect(const BBox& b) const{

//   float dist_squared = rayon * rayon;

//   if (position.getX() < b.min.getX()) 
//     dist_squared -= (position.getX() - b.min.getX())*(position.getX() - b.min.getX());
//   else if (position.getX() > b.max.getX())
//     dist_squared -= (position.getX() - b.max.getX())*(position.getX() - b.max.getX());
//   if (position.getY() < b.min.getY())
//     dist_squared -= (position.getY() - b.min.getY())*(position.getY() - b.min.getY());
//   else if (position.getY() > b.max.getY()) 
//     dist_squared -= (position.getY() - b.max.getY())*(position.getY() - b.max.getY());
//   if (position.getZ() < b.min.getZ())
//     dist_squared -= (position.getZ() - b.min.getZ())*(position.getZ() - b.min.getZ());
//   else if (position.getZ() > b.max.getZ())
//     dist_squared -= (position.getZ() - b.max.getZ())*(position.getZ() - b.max.getZ());
//   if(dist_squared > 0){
//     if(left != NULL || right != NULL){
//       return left->intersect(b) || right->intersect(b);	
//     }
//   }
//   return 0;
// }


// void gltDrawSphere(GLfloat fRadius, GLint iSlices, GLint iStacks)
// {
//   GLfloat drho = (GLfloat)(3.141592653589) / (GLfloat) iStacks;
//   GLfloat dtheta = 2.0f * (GLfloat)(3.141592653589) / (GLfloat) iSlices;
//   GLfloat ds = 1.0f / (GLfloat) iSlices;
//   GLfloat dt = 1.0f / (GLfloat) iStacks;
//   GLfloat t = 1.0f;	 
//   GLfloat s = 0.0f;
//   GLint i, j;     // Looping variables

//   for (i = 0; i < iStacks; i++) 
//     {
//       GLfloat rho = (GLfloat)i * drho;
//       GLfloat srho = (GLfloat)(sin(rho));
//       GLfloat crho = (GLfloat)(cos(rho));
//       GLfloat srhodrho = (GLfloat)(sin(rho + drho));
//       GLfloat crhodrho = (GLfloat)(cos(rho + drho));

//       // Many sources of OpenGL sphere drawing code uses a triangle fan
//       // for the caps of the sphere. This however introduces texturing 
//       // artifacts at the poles on some OpenGL implementations
//       glBegin(GL_TRIANGLE_STRIP);
//       s = 0.0f;
//       for ( j = 0; j <= iSlices; j++) 
// 	{
// 	  GLfloat theta = (j == iSlices) ? 0.0f : j * dtheta;
// 	  GLfloat stheta = (GLfloat)(-sin(theta));
// 	  GLfloat ctheta = (GLfloat)(cos(theta));

// 	  GLfloat x = stheta * srho;
// 	  GLfloat y = ctheta * srho;
// 	  GLfloat z = crho;
            
// 	  glTexCoord2f(s, t);
// 	  glNormal3f(x, y, z);
// 	  glVertex3f(x * fRadius, y * fRadius, z * fRadius);

// 	  x = stheta * srhodrho;
// 	  y = ctheta * srhodrho;
// 	  z = crhodrho;
// 	  glTexCoord2f(s, t - dt);
// 	  s += ds;
// 	  glNormal3f(x, y, z);
// 	  glVertex3f(x * fRadius, y * fRadius, z * fRadius);
// 	}
//       glEnd();

//       t -= dt;
//     }
// }

// void BSphere::draw() const{
//   glPushMatrix();
//   glTranslatef( position.getX(), position.getY(), position.getZ() );
//   gltDrawSphere(rayon, 8, 8);
//   glPopMatrix();
//   if(left != NULL)
//     left->draw();
//   // if(right != NULL)
//   //   right->draw();

// }

